/*-
 * Copyright (c) 2011 Author
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions, and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * TODO
 * 終了:
 * デバイスの認識 / 白凡/黒凡判定 / デバイス初期化 / LED表示
 * sysctl 雑多情報: desc / driver / location / pnpinfo / parent
 *
 * 今後の優先順位:
 * 周波数設定 / キャラクターデバイスの作成 / LNB 設定(要解析)
 *
 * recfriio コマンドからデバイスの操作方法を学習
 * 周波数設定 / 情報取得
 * AbstractFriio::open()
 * できたか? FriioBlack::setChannel() / FriioBlack::SetChannel_Black()
 * AbstractFriio::isStreamReady()
 * ストリームからのデータ取得処理の実装
 * USB_ERR_NO_PIPE は 1回の kldload(8) @ ufriio_attach() で4増える
 * ストリームデバイスファイル(character device: /dev/ptx0.t1)の生やし方の学習
 * AbstractFriio::startStream() @ AbstractFriio.cpp
 * AbstractFriio::getStream() @ AbstractFriio.cpp
 * AbstractFriio::stopStream() @ AbstractFriio.cpp
 * LNB トグル -> recfriio コマンドにその機能はない。
 *
 * 他ドライバーでのデータ操作手法
 * urio_ioctl() @ storage/urio.c 移植性を犠牲にしているため、
 * 完全な真似は出来ない。
 * serial/uplcom.c  uplcom_pl2303x_init pl2303x(配列)
 * 複数インターフェイスを持つデバイスのドライバー
 * net/uhso.c serial/ufoma.c storase/umass.c
 * uhso_attach() -> control_config の初期化 -> uhso_probe_iface()
 * -> uhso_attach_bulkserial()
 * uhso_detach() -> usbd_transfer_unsetup() をまとめて実行
 * ufoma_attach() -> ufma_modem_setup() for ループ -> usbd_set_parent_iface()
 * -> usbd_transfer_setup()
 *
 * bulk 転送と usb_fifo_attach が絡むのは serial/ulpt.c / storage/urio.c のみ
 * sysctl(3) インターフェイスの学習
 * sys/dev/usb/input/ums.c
 * sys/netgraph/bluetooth/drivers/ubt/ng_ubt.c
 * dev.ptx.* のインターフェイスの移植
 * ptx_sysctl_init() @ ptx_sysctl.c
 * pt_make_tuner() @ ptx_tuner.c
 * sysctl 必須情報: lnb / [st]0.{freq,signal}	(s1 とか t1 はいらない)
 * 最終: コード / ライセンス文のクリーンアップ
 */
#include <sys/cdefs.h>
__FBSDID("$FreeBSD$");


#include <sys/param.h>
//#include <sys/stdint.h>
#include <sys/stddef.h>
//#include <sys/queue.h>
#include <sys/types.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/bus.h>
//#include <sys/linker_set.h>
#include <sys/module.h>
#include <sys/lock.h>
#include <sys/mutex.h>
#include <sys/sysctl.h>
//#include <sys/condvar.h>
//#include <sys/sx.h>
//#include <sys/unistd.h>
//#include <sys/callout.h>
//#include <sys/malloc.h>
//#include <sys/priv.h>
#include <sys/conf.h>	// UID_ROOT, GID_OPERATOR
#include <sys/fcntl.h>

#include <dev/usb/usb.h>
#include <dev/usb/usbdi.h>
#include "usbdevs.h"

#define	USB_DEBUG_VAR usb_debug
#include <dev/usb/usb_debug.h>

//#include <dev/usb/ufriio_ioctl.h>

SYSCTL_NODE(_hw_usb, OID_AUTO, ufriio, CTLFLAG_RW, 0, "Friio");

#define UFRIIO_LED_PURPLE	0x1	/* waiting */
#define UFRIIO_LED_GREEN	0x2	/* watching */
#define UFRIIO_LED_RED		0x3	/* recording */
#define UFRIIO_LED_WHITE	0x4	/* error */

#define	UFRIIO_TUNER_TYPE_ERROR	0x0
#define	UFRIIO_TUNER_TYPE_WHITE	0x1
#define	UFRIIO_TUNER_TYPE_BLACK	0x2

#define	TS_DATASIZE		16384
#define	UFRIIO_BUFFER_TIME	10UL
// FIXME 白凡・黒凡でサイズが変わる?
#define	UFRIIO_FIFO_BUFSIZE	(0x200000UL * UFRIIO_BUFFER_TIME) / TS_DATASIZE
// 0xbc00 か?
//#define	UFRIIO_FIFO_BUFSIZE	(0x400000UL * UFRIIO_BUFFER_TIME) / TS_DATASIZE;    // 黒凡

#define USB_VENDOR_FRIIO	0x7a69
#define USB_PRODUCT_FRIIO	0x0001

enum {
	// FIXME need = 0
	UFRIIO_BULK_XFER_READ,
	UFRIIO_BULK_XFER_WRITE,
	UFRIIO_BULK_XFER_MAX,	/* 2 */
};
enum {
	UFRIIO_INTR_XFER = 0,
	UFRIIO_INTR_XFER_MAX,	/* 1 */
};

struct ufriio_softc {
	struct usb_fifo_sc sc_fifo;
	struct mtx sc_mtx;
	struct usb_xfer *sc_bulk_xfer[UFRIIO_BULK_XFER_MAX];
	//struct usb_xfer *sc_intr_xfer[UFRIIO_INTR_XFER_MAX];

	struct usb_device *sc_udev;

	uint32_t sc_unit;

	uint32_t sc_channel;
	uint32_t sc_tuner_type;
	uint32_t sc_lnb;

	uint8_t	sc_name[16];
	uint8_t sc_intr_iface_no;
	uint8_t sc_intr_iface_index;
	uint8_t sc_bulk_iface_index;
};

struct ufriio_black_channel_sequence {
	uint8_t req_no;
	uint16_t value;
	uint16_t index;
	uint8_t data_exist;
	uint8_t data1;
	uint8_t data2;
	uint8_t data3;
} ufriio_black_commands[] = {
								// Index
	{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x00, 0x0A },	// 0
	{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x01, 0x4C },	// 1
	{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x02, 0x02 },	// 2
	{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x03, 0x40 },	// 3
	{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x04, 0x6C },	// 4
	{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x05, 0x47 },	// 5
	{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x06, 0x02 },	// 6
	{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x07, 0x30 },	// 7
	{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x08, 0x20 },	// 8
	{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x09, 0xA2 },	// 9
	{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x0A, 0xF1 },	// 10
	{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x0B, 0x38 },	// 11
	{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x0C, 0xD0 },	// 12
	{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x0D, 0x10 },	// 13
	{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x0E, 0xB0 },	// 14
	{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x0F, 0x01 },	// 15
	{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x00, 0x0A },	// 16
	{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x01, 0x4C },	// 17
	{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x02, 0x02 },	// 18
	{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x03, 0x40 },	// 19
	{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x04, 0x6C },	// 20
	{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x05, 0x47 },	// 21
	{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x06, 0x02 },	// 22
	{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x07, 0x30 },	// 23
	{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x08, 0x20 },	// 24
	{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x09, 0xA2 },	// 25
	{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x0A, 0xF1 },	// 26
	{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x0B, 0x38 },	// 27
	{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x0C, 0xD0 },	// 28
	{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x0D, 0x10 },	// 29
	{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x0E, 0xB0 },	// 30
	{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x0F, 0x01 },	// 31
	{ 0x01, 0x3210, 0x0001, FALSE, 0x00, 0x00, 0x00 },	// 32
	{ 0x01, 0x3200, 0x0002, FALSE, 0x00, 0x00, 0x00 },	// 33
	{ 0x01, 0x3200, 0x0003, FALSE, 0x00, 0x00, 0x00 },	// 34
	{ 0x01, 0x3200, 0x0008, FALSE, 0x00, 0x00, 0x00 },	// 35
	{ 0x01, 0x3259, 0x000C, FALSE, 0x00, 0x00, 0x00 },	// 36
	{ 0x01, 0x32F2, 0x000D, FALSE, 0x00, 0x00, 0x00 },	// 37
	{ 0x01, 0x3251, 0x000F, FALSE, 0x00, 0x00, 0x00 },	// 38
	{ 0x01, 0x32B2, 0x0010, FALSE, 0x00, 0x00, 0x00 },	// 39
	{ 0x01, 0x3204, 0x001A, FALSE, 0x00, 0x00, 0x00 },	// 40
	{ 0x01, 0x3204, 0x001B, FALSE, 0x00, 0x00, 0x00 },	// 41
	{ 0x01, 0x3240, 0x0038, FALSE, 0x00, 0x00, 0x00 },	// 42
	{ 0x01, 0x3210, 0x0039, FALSE, 0x00, 0x00, 0x00 },	// 43
	{ 0x01, 0x3208, 0x003C, FALSE, 0x00, 0x00, 0x00 },	// 44
	{ 0x01, 0x3296, 0x003D, FALSE, 0x00, 0x00, 0x00 },	// 45
	{ 0x01, 0x32D9, 0x003E, FALSE, 0x00, 0x00, 0x00 },	// 46
	{ 0x01, 0x32B6, 0x003F, FALSE, 0x00, 0x00, 0x00 },	// 47
	{ 0x01, 0x325D, 0x0040, FALSE, 0x00, 0x00, 0x00 },	// 48
	{ 0x01, 0x3299, 0x0041, FALSE, 0x00, 0x00, 0x00 },	// 49
	{ 0x01, 0x3258, 0x0044, FALSE, 0x00, 0x00, 0x00 },	// 50
	{ 0x01, 0x3254, 0x0045, FALSE, 0x00, 0x00, 0x00 },	// 51
	{ 0x01, 0x3230, 0x0046, FALSE, 0x00, 0x00, 0x00 },	// 52
	{ 0x01, 0x3241, 0x0047, FALSE, 0x00, 0x00, 0x00 },	// 53
	{ 0x01, 0x3216, 0x0049, FALSE, 0x00, 0x00, 0x00 },	// 54
	{ 0x01, 0x3241, 0x004A, FALSE, 0x00, 0x00, 0x00 },	// 55
	{ 0x01, 0x3241, 0x004B, FALSE, 0x00, 0x00, 0x00 },	// 56
	{ 0x01, 0x325E, 0x004F, FALSE, 0x00, 0x00, 0x00 },	// 57
	{ 0x01, 0x3204, 0x0050, FALSE, 0x00, 0x00, 0x00 },	// 58
	{ 0x01, 0x32C0, 0x0051, FALSE, 0x00, 0x00, 0x00 },	// 59
	{ 0x01, 0x328A, 0x0052, FALSE, 0x00, 0x00, 0x00 },	// 60
	{ 0x01, 0x3213, 0x0053, FALSE, 0x00, 0x00, 0x00 },	// 61
	{ 0x01, 0x3290, 0x0054, FALSE, 0x00, 0x00, 0x00 },	// 62
	{ 0x01, 0x3292, 0x0055, FALSE, 0x00, 0x00, 0x00 },	// 63
	{ 0x01, 0x326D, 0x0058, FALSE, 0x00, 0x00, 0x00 },	// 64
	{ 0x01, 0x326F, 0x0059, FALSE, 0x00, 0x00, 0x00 },	// 65
	{ 0x01, 0x322E, 0x005A, FALSE, 0x00, 0x00, 0x00 },	// 66
	{ 0x01, 0x3223, 0x005B, FALSE, 0x00, 0x00, 0x00 },	// 67
	{ 0x01, 0x3233, 0x005D, FALSE, 0x00, 0x00, 0x00 },	// 68
	{ 0x01, 0x3204, 0x0087, FALSE, 0x00, 0x00, 0x00 },	// 69
	{ 0x01, 0x3200, 0x008D, FALSE, 0x00, 0x00, 0x00 },	// 70
	{ 0x01, 0x3000, 0x0077, FALSE, 0x00, 0x00, 0x00 },	// 71
	{ 0x01, 0x3000, 0x0071, FALSE, 0x00, 0x00, 0x00 },	// 72
	{ 0x01, 0x3000, 0x0071, FALSE, 0x00, 0x00, 0x00 },	// 73
	{ 0x01, 0x3240, 0x008F, FALSE, 0x00, 0x00, 0x00 },	// 74
	{ 0x01, 0x32F2, 0x0090, FALSE, 0x00, 0x00, 0x00 },	// 75
	{ 0x01, 0x3207, 0x00A5, FALSE, 0x00, 0x00, 0x00 },	// 76
	{ 0x01, 0x3207, 0x00A5, FALSE, 0x00, 0x00, 0x00 },	// 77
	{ 0x01, 0x32FF, 0x00A3, FALSE, 0x00, 0x00, 0x00 },	// 78
	{ 0x01, 0x32F7, 0x00A3, FALSE, 0x00, 0x00, 0x00 },	// 79
	{ 0x01, 0x3040, 0x0001, FALSE, 0x00, 0x00, 0x00 },	// 80
	{ 0x01, 0x3030, 0x0004, FALSE, 0x00, 0x00, 0x00 },	// 81
	{ 0x01, 0x3040, 0x0005, FALSE, 0x00, 0x00, 0x00 },	// 82
	{ 0x01, 0x3040, 0x0007, FALSE, 0x00, 0x00, 0x00 },	// 83
	{ 0x01, 0x30FF, 0x000F, FALSE, 0x00, 0x00, 0x00 },	// 84
	{ 0x01, 0x3027, 0x0011, FALSE, 0x00, 0x00, 0x00 },	// 85
	{ 0x01, 0x302F, 0x0013, FALSE, 0x00, 0x00, 0x00 },	// 86
	{ 0x01, 0x3031, 0x0014, FALSE, 0x00, 0x00, 0x00 },	// 87
	{ 0x01, 0x3002, 0x0016, FALSE, 0x00, 0x00, 0x00 },	// 88
	{ 0x01, 0x3020, 0x0022, FALSE, 0x00, 0x00, 0x00 },	// 89
	{ 0x01, 0x3070, 0x0024, FALSE, 0x00, 0x00, 0x00 },	// 90
	{ 0x01, 0x3079, 0x002C, FALSE, 0x00, 0x00, 0x00 },	// 91
	{ 0x01, 0x3034, 0x002D, FALSE, 0x00, 0x00, 0x00 },	// 92
	{ 0x01, 0x3028, 0x0030, FALSE, 0x00, 0x00, 0x00 },	// 93
	{ 0x01, 0x3031, 0x0031, FALSE, 0x00, 0x00, 0x00 },	// 94
	{ 0x01, 0x3001, 0x0038, FALSE, 0x00, 0x00, 0x00 },	// 95
	{ 0x01, 0x3078, 0x0039, FALSE, 0x00, 0x00, 0x00 },	// 96
	{ 0x01, 0x3090, 0x0048, FALSE, 0x00, 0x00, 0x00 },	// 97
	{ 0x01, 0x3068, 0x0051, FALSE, 0x00, 0x00, 0x00 },	// 98
	{ 0x01, 0x3038, 0x005E, FALSE, 0x00, 0x00, 0x00 },	// 99
	{ 0x01, 0x3021, 0x00C0, FALSE, 0x00, 0x00, 0x00 },	// 100
	{ 0x01, 0x3010, 0x00C1, FALSE, 0x00, 0x00, 0x00 },	// 101
	{ 0x01, 0x301A, 0x00E4, FALSE, 0x00, 0x00, 0x00 },	// 102
	{ 0x01, 0x301F, 0x00EA, FALSE, 0x00, 0x00, 0x00 },	// 103
	{ 0x00, 0x0000, 0x0000, FALSE, 0x00, 0x00, 0x00 },	// End
};

/* prototypes */
static device_probe_t ufriio_probe;
static device_attach_t ufriio_attach;
static device_detach_t ufriio_detach;

static usb_fifo_open_t ufriio_open;
static usb_fifo_close_t ufriio_close;
static usb_fifo_cmd_t ufriio_start_read;
static usb_fifo_cmd_t ufriio_stop_read;
//static usb_fifo_ioctl_t ufriio_ioctl;

static usb_callback_t ufriio_read_callback;
static usb_callback_t ufriio_write_callback;

// FIXME ses also: PushIoThread::operator()()
static const struct usb_config ufriio_bulk_config[UFRIIO_BULK_XFER_MAX] = {
	[UFRIIO_BULK_XFER_READ] = {
		.type = UE_BULK,
		.endpoint = 0x81,
		//.endpoint = 0x01,
		//.endpoint = UE_ADDR_ANY,
		.direction = UE_DIR_IN,
		//.interval = 400,
		//.timeout =  2000,
		.timeout =  200,
		//.bufsize = 0x0200,
		.bufsize = TS_DATASIZE,
		//.flags = {.pipe_bof = 1,.force_short_xfer = 1,.proxy_buffer = 1,},
		.flags = {.pipe_bof = 1,.short_xfer_ok = 1,},
		//.flags = {.pipe_bof = 1,.force_short_xfer = 1,},
		//.flags = {.proxy_buffer = 1,.short_xfer_ok = 1,},
		//.flags = {.no_pipe_ok=1,},
		//.flags = {.pipe_bof=1,},
		//.flags = {.pipe_bof=1, .proxy_buffer = 1},
		// note: usbd_transfer_start() のマニュアルの
		// セクション USB TRANSFER CALLBACK
		.callback = &ufriio_read_callback,
	},
	[UFRIIO_BULK_XFER_WRITE] = {
		.type = UE_BULK,
		.endpoint = 0x02,
		.direction = UE_DIR_OUT,
		//.bufsize = 0x0200,
		.bufsize = TS_DATASIZE,
		//.flags = {.pipe_bof=1, .short_xfer_ok=1,},
		// FIXME is it corrent?
		.callback = &ufriio_write_callback,
	},
};
/*
static const struct usb_config ufriio_intr_config[UFRIIO_INTR_XFER_MAX] = {
	[UFRIIO_INTR_XFER] = {
		.type = UE_INTERRUPT,
		.endpoint = 0x83,
		.direction = UE_DIR_IN,
		.bufsize = 0x0040,
		.flags = {.pipe_bof=1, .short_xfer_ok=1,},
		.callback = &ufriio_read_callback,
	},
};
*/

static struct usb_fifo_methods ufriio_fifo_methods = {
	.f_open       = &ufriio_open,
	.f_close      = &ufriio_close,
	// FIXME ufriio_start_read() がそもそも実行されていない?
	.f_start_read = &ufriio_start_read,
	.f_stop_read  = &ufriio_stop_read,
	//.f_ioctl = &ufriio_ioctl,
	.basename[0] = "ufriio",
};

static int	ufriio_read_request_1(struct ufriio_softc *, uint8_t, uint16_t,
		    uint16_t, uint8_t *);
static int	ufriio_read_request_n(struct ufriio_softc *, uint8_t, uint16_t,
		    uint16_t, uint8_t *, size_t);
static int	ufriio_write_request_0(struct ufriio_softc *, uint8_t, uint16_t,
		    uint16_t);
static int	ufriio_write_request_n(struct ufriio_softc *, uint8_t, uint16_t,
		    uint16_t, uint8_t *, size_t);
int		ufriio_sysctl_channel(SYSCTL_HANDLER_ARGS);
int		ufriio_sysctl_signal_level(SYSCTL_HANDLER_ARGS);

void ufriio_stream_init(struct ufriio_softc *);

void ufriio_black_initialize(struct ufriio_softc *);
int ufriio_black_set_channel(struct ufriio_softc *, uint32_t, uint32_t);
int ufriio_black_set_channel_internal(struct ufriio_softc *, uint32_t, uint32_t);
int ufriio_black_get_signal_level(struct ufriio_softc *);

void ufriio_white_initialize(struct ufriio_softc *);
int ufriio_white_set_channel(struct ufriio_softc *, uint32_t);
int ufriio_white_get_signal_level(struct ufriio_softc *);

void ufriio_set_led_color(struct ufriio_softc *, int);

int ufriio_get_tuner_type(struct ufriio_softc *);


static devclass_t ufriio_devclass;

static device_method_t ufriio_methods[] = {
	DEVMETHOD(device_probe, ufriio_probe),
	DEVMETHOD(device_attach, ufriio_attach),
	DEVMETHOD(device_detach, ufriio_detach),
	{0, 0}
};

static driver_t ufriio_driver = {
	.name = "ufriio",
	.methods = ufriio_methods,
	.size = sizeof(struct ufriio_softc),
};

DRIVER_MODULE(ufriio, uhub, ufriio_driver, ufriio_devclass, NULL, 0);
MODULE_DEPEND(ufriio, usb, 1, 1, 1);
MODULE_VERSION(ufriio, 1);

static int
ufriio_probe(device_t dev)
{
	struct usb_attach_arg *uaa = device_get_ivars(dev);

	if (uaa->usb_mode != USB_MODE_HOST) {
		return (ENXIO);
	}
	if ((uaa->info.idVendor == USB_VENDOR_FRIIO) &&
	    (uaa->info.idProduct == USB_PRODUCT_FRIIO) &&
	    (uaa->info.bIfaceIndex == 0)) {
	    // FIXME これで合っているか?
	    //(uaa->info.bInterfaceClass == UICLASS_HID)) {
		return (0);
	}
	return (ENXIO);
}

static int
ufriio_attach(device_t dev)
{
	struct usb_attach_arg *uaa = device_get_ivars(dev);
	struct ufriio_softc *sc = device_get_softc(dev);
	int tuner_type;
	int error;

	sc->sc_udev = uaa->device;
	sc->sc_unit = device_get_unit(dev);

	snprintf(sc->sc_name, sizeof(sc->sc_name), "%s",
	    device_get_nameunit(dev));

	// device_set_usb_desc() call before / after mtx_init ?
	device_set_usb_desc(dev);

	mtx_init(&sc->sc_mtx, "ufriio lock", NULL, MTX_DEF | MTX_RECURSE);

	// FIXME 順序はここでいいのか?
	// 固定処理A / friio_initialize()
	tuner_type = ufriio_get_tuner_type(sc);
	switch (tuner_type) {
	case UFRIIO_TUNER_TYPE_WHITE:
		printf("%s: white(ISDB-T)\n", device_get_nameunit(dev));
		break;
	case UFRIIO_TUNER_TYPE_BLACK:
		printf("%s: black(ISDB-S)\n", device_get_nameunit(dev));
		break;
	default:
		goto detach;
		break;
	}

	// FIXME 初期化処理が足りなかった?
	// FIXME この処理は ufriio_attach ではなく、ufriio_start_read() か? 逆か?
	// dvb ドライバーでは、 restart: ラベルの後に置かれている。
	if (sc->sc_tuner_type == UFRIIO_TUNER_TYPE_WHITE) {
		ufriio_white_initialize(sc);
	} else {
		ufriio_black_initialize(sc);
	}

	/* Get interrupt interface */
	/*
	// FIXME このコードは不要?
	sc->sc_intr_iface_index = 1;
	error = usbd_transfer_setup(uaa->device, &sc->sc_intr_iface_index,
	    sc->sc_intr_xfer, ufriio_intr_config, UFRIIO_INTR_XFER_MAX,
	    sc, &sc->sc_mtx);
	if (error) {
		DPRINTF("error=%s\n", usbd_errstr(error));
		printf("error=%s\n", usbd_errstr(error));//FIXME
		goto detach;
	}
	*/

	/* Get bulk transfer interface */
	//sc->sc_bulk_iface_index = 0;
	sc->sc_bulk_iface_index = uaa->info.bIfaceIndex;
	error = usbd_set_alt_interface_index(uaa->device,
	    sc->sc_bulk_iface_index, 0);
	if (error) {
		DPRINTF("error=%s\n", usbd_errstr(error));
		printf("error=%s\n", usbd_errstr(error));//FIXME
		goto detach;
	}

	error = usbd_transfer_setup(uaa->device, &sc->sc_bulk_iface_index,
	    sc->sc_bulk_xfer, ufriio_bulk_config, UFRIIO_BULK_XFER_MAX,
	    sc, &sc->sc_mtx);
	    //sc->sc_bulk_xfer, ufriio_bulk_config, UFRIIO_BULK_XFER_MAX,
	if (error) {
		DPRINTF("error=%s\n", usbd_errstr(error));
		printf("error=%s\n", usbd_errstr(error));//FIXME
		goto detach;
	}

	// FIXME ここではファイルのオーナー/グループが設定される。
	error = usb_fifo_attach(uaa->device, sc, &sc->sc_mtx,
	    &ufriio_fifo_methods, &sc->sc_fifo,
	    //device_get_unit(dev), 0, sc->sc_bulk_iface_index,
	    device_get_unit(dev), 0 - 1, sc->sc_bulk_iface_index,
	    UID_ROOT, GID_OPERATOR, 0666);
	if (error) {
		printf("usb_fifo_attach: failed\n"); //FIXME
		goto detach;
	}

	SYSCTL_ADD_PROC(device_get_sysctl_ctx(dev),
	    SYSCTL_CHILDREN(device_get_sysctl_tree(dev)),
	    OID_AUTO, "channel", CTLTYPE_INT|CTLFLAG_RW|CTLFLAG_ANYBODY,
	    sc, 0, ufriio_sysctl_channel,
	    "I", "Channel number");

	SYSCTL_ADD_PROC(device_get_sysctl_ctx(dev),
	    SYSCTL_CHILDREN(device_get_sysctl_tree(dev)),
	    OID_AUTO, "slevel", CTLTYPE_INT|CTLFLAG_RD,
	    sc, 0, ufriio_sysctl_signal_level,
	    "I", "signal level");

	return (0);			/* success */

detach:
	ufriio_detach(dev);
	return (ENXIO);
}

static int
ufriio_detach(device_t dev)
{
	struct ufriio_softc *sc = device_get_softc(dev);

	usb_fifo_detach(&sc->sc_fifo);

	usbd_transfer_unsetup(sc->sc_bulk_xfer, UFRIIO_BULK_XFER_MAX);
	//usbd_transfer_unsetup(sc->sc_intr_xfer, UFRIIO_INTR_XFER_MAX);

	mtx_destroy(&sc->sc_mtx);

	return (0);
}

int
ufriio_sysctl_channel(SYSCTL_HANDLER_ARGS)
{
	struct ufriio_softc *sc = (struct ufriio_softc *)oidp->oid_arg1;
	int error;

	error = sysctl_handle_int(oidp, &sc->sc_channel, 0, req);
	if (error != 0 || req->newptr == NULL) {
		return (error);
	}

	if (sc->sc_tuner_type == UFRIIO_TUNER_TYPE_WHITE) {
		return ufriio_white_set_channel(sc, sc->sc_channel);
	} else {
		// FIXME 1(UFRIIO_BS) は仮の値
		return ufriio_black_set_channel(sc, 1, sc->sc_channel);
	}
}

int
ufriio_sysctl_signal_level(SYSCTL_HANDLER_ARGS)
{
	struct ufriio_softc *sc = (struct ufriio_softc *)oidp->oid_arg1;
	int signal;

	if (sc->sc_tuner_type == UFRIIO_TUNER_TYPE_WHITE) {
		signal = ufriio_white_get_signal_level(sc);
	} else {
		signal = ufriio_black_get_signal_level(sc);
	}
	return sysctl_handle_int(oidp, &signal, 0, req);
}

// AbstractFriio::UsbInitalize() 相当
static int
ufriio_open(struct usb_fifo *fifo, int fflags)
{
	struct ufriio_softc *sc = usb_fifo_softc(fifo);
	// ptxopen() @ ptx_tuner.c のようなコーディングか?
	//
	// write モードで開くことはない?
	
	// FIXME デバッグの邪魔
	ufriio_set_led_color(sc, UFRIIO_LED_GREEN);

	if (fflags & FREAD) {
		// FIXME
		// 転送サイズ設定と言う処理が抜けているのか?

		// FIXME サイズ未確定
		// PushIoThread::operator() @ IoThread.cpp
		// TS_DATASIZE(16384) は Linux の usbfs 依存?
		// TSパケットは 188 で、その倍数ではない。87パケット分。
		// PopIoThread::operator()() {
		//if (usb_fifo_alloc_buffer(fifo, TS_DATASIZE, UFRIIO_FIFO_BUFSIZE)) {
		if (usb_fifo_alloc_buffer(fifo, 188, UFRIIO_FIFO_BUFSIZE)) {
			return (ENOMEM);
		}
	}

	return (0);
}

static void
ufriio_close(struct usb_fifo *fifo, int fflags)
{
	struct ufriio_softc *sc = usb_fifo_softc(fifo);
	//FIXME デバッグの邪魔
	ufriio_set_led_color(sc, UFRIIO_LED_PURPLE);

	// ptxclose() @ ptx_tuner.c のようなコーディングか?
	// sys/dev/usb/serial/ulpt.c 
	// のパクリだが、みんな似たようなことをやっている?

	// FREAD / FWRITE フラグをリセット
	// ufriio では、FREAD だけで十分? これもいらない?
	//sc->sc_fflags &= ~(fflags & (FREAD | FWRITE));

	if (fflags & (FREAD | FWRITE)) {
		usb_fifo_free_buffer(fifo);
	}

	// stopStream()
	//
	// UsbProcLED(purple)
	if (sc->sc_tuner_type == UFRIIO_TUNER_TYPE_WHITE) {
		// ISDB-T mode
		// UsbProcEnd()
		ufriio_write_request_0(sc, 0x01, 0x0001, 0x0000);
	} else {
		// ISDB-S mode
		// UsbProcEnd()
		ufriio_write_request_0(sc, 0x01, 0x0000, 0x0000);
	}
}

static void
ufriio_start_read(struct usb_fifo *fifo)
{
	// FIXME
	// 周波数があっていない場合に、recfriio はどのようなデータを
	// 出力するかを知る必要がある。0バイトか?ゴミデータか?
	struct ufriio_softc *sc = usb_fifo_softc(fifo);

	DPRINTF("ufriio_start_read(usb_fifo 0x%x)\n", fifo);
	printf("ufriio_start_read(usb_fifo 0x%p)\n", fifo);//FIXME

	// 解析ログの「チューナー制御開始」とは?
	// from storage/urio.c
	// FIXME
	usbd_transfer_start(sc->sc_bulk_xfer[UFRIIO_BULK_XFER_READ]);
	//usbd_transfer_start(sc->sc_bulk_xfer[UFRIIO_BULK_XFER_WRITE]);
}

static void
ufriio_stop_read(struct usb_fifo *fifo)
{
	// FIXME
	struct ufriio_softc *sc = usb_fifo_softc(fifo);

	DPRINTFN(1, "ufriio_stop_read(usb_fifo 0x%x)\n", fifo);
	printf("ufriio_stop_read(usb_fifo 0x%p)\n", fifo);//FIXME

	usbd_transfer_stop(sc->sc_bulk_xfer[UFRIIO_BULK_XFER_READ]);
	//usbd_transfer_stop(sc->sc_bulk_xfer[UFRIIO_BULK_XFER_WRITE]);
}

// misc/udbp.c: udbp_bulk_read_callback
// input/uhid.c: uhid_intr_callback (not uhid_read_callback)
// 
static void
ufriio_read_callback(struct usb_xfer *xfer, usb_error_t error)
{
	struct ufriio_softc *sc = usbd_xfer_softc(xfer);
	struct usb_fifo *f = sc->sc_fifo.fp[USB_FIFO_RX];
	struct usb_page_cache *pc;
	int actlen;

	usbd_xfer_status(xfer, &actlen, NULL, NULL, NULL);

	if (f == NULL) {
		// FIXME
		printf("Error: No FIFO\n");
		return;
	}
	DPRINTF("state=0x%x\n", USB_GET_STATE(xfer));

	switch (USB_GET_STATE(xfer)) {
	case USB_ST_TRANSFERRED:
		printf("ufriio_read_callback(): transferred.\n");   // FIXME

		pc = usbd_xfer_get_frame(xfer, 0);
		// usbd_copy_out() はこの関数内にある。
		usb_fifo_put_data(f, pc, 0, actlen, 1);
		break;
	case USB_ST_SETUP:
		// FIXME なにかしらのリクエストを叩かないといけないか?

		printf("ufriio_read_callback(): setup.\n");   // FIXME

		// usbd_xfer_set_priv()
		// usbd_copy_in() または usbd_xfer_set_frame_data()
		// frame_data() には usbd_xfer_set_frame_len() が入る。 

		// FIXME usb_fifo_put_bytes_max()
		//if (usb_fifo_put_bytes_max(f) != 0) {
setup_next:
		//if (usb_fifo_put_bytes_max(f) > 0) {
			//printf("usbd_xfer_max_len(): %d\n", usbd_xfer_max_len(xfer));//FIXME
			//usbd_xfer_set_frames(xfer, 1);
			usbd_xfer_set_frame_len(xfer, 0, usbd_xfer_max_len(xfer));
			usbd_transfer_submit(xfer);
		//}
		// usbd_xfer_get_frame()
		// usb_fifo_put_data_linear()
		break;
	default:
		// FIXME このエラーは実行開始時点でこうなっている!
		// これはエラーじゃない。これも通る道。

		printf("ufriio_read_callback(): default.\n");   // FIXME
		if (error != USB_ERR_CANCELLED) {
			usbd_xfer_set_stall(xfer);
			goto setup_next;
		}
		break;
	}

	return;
}

 // FIXME
static void
ufriio_write_callback(struct usb_xfer *xfer, usb_error_t error)
{
	struct ufriio_softc *sc = usbd_xfer_softc(xfer);
	struct usb_fifo *f = sc->sc_fifo.fp[USB_FIFO_TX];
	struct usb_page_cache *pc;
	int actlen, max;

	usbd_xfer_status(xfer, &actlen, NULL, NULL, NULL);

	switch (USB_GET_STATE(xfer)) {
	case USB_ST_TRANSFERRED:
		printf("ufriio_write_callback(): transferred.\n");   // FIXME
	case USB_ST_SETUP:
		printf("ufriio_write_callback(): setup.\n");   // FIXME
		pc = usbd_xfer_get_frame(xfer, 0);
		max = usbd_xfer_max_len(xfer);

		if (usb_fifo_get_data(f, pc, 0, max, &actlen, 0)) {
			usbd_xfer_set_frame_len(xfer, 0, actlen);
			usbd_transfer_submit(xfer);
		}
		break;
	default: 
		printf("ufriio_write_callback(): default.\n");   // FIXME
		if (error != USB_ERR_CANCELLED) {
			usbd_xfer_set_stall(xfer);
		}
		break;
	}

	return;
}


// ^ public
//************************************************************************
// FIXME Internal

// 「コントロール転送」処理を行う
static int
ufriio_read_request_1(struct ufriio_softc *sc, uint8_t request,
    uint16_t value, uint16_t index, uint8_t *retbuf)
{
	struct usb_device_request req;
	int error;
	uint8_t buf[1];

	// FIXME  (UT_READ  | UT_VENDOR | UT_DEVICE) @ usb.h
	// Linux の USB_DIR_IN | USB_TYPE_VENDOR | UT_DEVICE と同じ
	// USB_DIR_OUT ... と同じ働きをするのは、 UT_WRITE_VENDOR_DEVICE
	// serial/ufoma.c, controller/uss820dci.c
	req.bmRequestType = UT_READ_VENDOR_DEVICE;
	req.bRequest = request;
	USETW(req.wValue, value);
	USETW(req.wIndex, index);
	USETW(req.wLength, 1);

	error = usbd_do_request(sc->sc_udev, NULL, &req, buf);

	if (retbuf) {
		*retbuf = buf[0];
	}
	if (error) {
		return (ENXIO);
	}
	return (0);
}

static int
ufriio_read_request_n(struct ufriio_softc *sc, uint8_t request,
    uint16_t value, uint16_t index, uint8_t *retbuf, size_t length)
{
	struct usb_device_request req;
	int error;

	// FIXME  (UT_READ  | UT_VENDOR | UT_DEVICE) @ usb.h
	// Linux の USB_DIR_IN | USB_TYPE_VENDOR | UT_DEVICE と同じ
	// USB_DIR_OUT ... と同じ働きをするのは、 UT_WRITE_VENDOR_DEVICE
	// serial/ufoma.c, controller/uss820dci.c
	req.bmRequestType = UT_READ_VENDOR_DEVICE;
	req.bRequest = request;
	USETW(req.wValue, value);
	USETW(req.wIndex, index);
	USETW(req.wLength, length);

	error = usbd_do_request(sc->sc_udev, NULL, &req, retbuf);

	if (error) {
		return (ENXIO);
	}
	return (0);
}

static int
ufriio_write_request_0(struct ufriio_softc *sc, uint8_t request,
    uint16_t value, uint16_t index)
{
	struct usb_device_request req;
	int error;

	req.bmRequestType = UT_WRITE_VENDOR_DEVICE;
	req.bRequest = request;
	USETW(req.wValue, value);
	USETW(req.wIndex, index);
	USETW(req.wLength, 0);

	error = usbd_do_request(sc->sc_udev, NULL, &req, NULL);

	if (error) {
		return (ENXIO);
	}
	return (0);
}

static int
ufriio_write_request_n(struct ufriio_softc *sc, uint8_t request,
    uint16_t value, uint16_t index, uint8_t *data, size_t length)
{
	struct usb_device_request req;
	int error;

	// FIXME  (UT_READ  | UT_VENDOR | UT_DEVICE) @ usb.h
	// Linux の USB_DIR_IN | USB_TYPE_VENDOR | UT_DEVICE と同じ
	// USB_DIR_OUT ... と同じ働きをするのは、 UT_WRITE_VENDOR_DEVICE
	// serial/ufoma.c, controller/uss820dci.c
	req.bmRequestType = UT_WRITE_VENDOR_DEVICE;
	req.bRequest = request;
	USETW(req.wValue, value);
	USETW(req.wIndex, index);
	USETW(req.wLength, length);

	error = usbd_do_request(sc->sc_udev, NULL, &req, data);

	if (error) {
		return (ENXIO);
	}
	return (0);
}

/* Tuner type checker */
// 固定処理A
int
ufriio_get_tuner_type(struct ufriio_softc *sc)
{
	int tuner_type = UFRIIO_TUNER_TYPE_ERROR;

	// STEP 1
	ufriio_write_request_0(sc, 0x01, 0x0002, 0x0011);
	// FIXME ufriio_write_request_x() -> gl861_i2c_msg()

	// STEP 2
	ufriio_write_request_0(sc, 0x01, 0x0000, 0x0011);

	// FIXME ここからループ?
	// STEP 3
	uint8_t input[] = "\xc1";
	ufriio_write_request_n(sc, 0x03, 0x3000, 0x00fe, input, 1);

	// STEP 4
	int8_t resultbuf1[] = "";
	ufriio_read_request_1(sc, 0x02, 0x3000, 0x0100, resultbuf1);

	if (resultbuf1[0]) {
		// ISDB-T
		tuner_type = UFRIIO_TUNER_TYPE_WHITE;
	} else {
		// ISDB-S
		tuner_type = UFRIIO_TUNER_TYPE_BLACK;

		uint8_t resultbuf2[] = "";
		ufriio_read_request_1(sc, 0x02, 0x3000, 0x0080, resultbuf2);
	}
	// FIXME ここまでループ?

	// STEP 5
	ufriio_write_request_0(sc, 0x01, 0x0004, 0x0030);

	sc->sc_tuner_type = tuner_type;
	return (tuner_type);
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// FIXME
// 各チューナー共通処理
void
ufriio_stream_init(struct ufriio_softc *sc)
{
	ufriio_write_request_0(sc, 0x01, 0x0008, 0x0033);
	ufriio_write_request_0(sc, 0x01, 0x0040, 0x0037);
	ufriio_write_request_0(sc, 0x01, 0x001f, 0x003a);
	ufriio_write_request_0(sc, 0x01, 0x00ff, 0x003b);
	ufriio_write_request_0(sc, 0x01, 0x001f, 0x003c);
	ufriio_write_request_0(sc, 0x01, 0x00ff, 0x003d);
	ufriio_write_request_0(sc, 0x01, 0x0000, 0x0038);
	ufriio_write_request_0(sc, 0x01, 0x0000, 0x0035);
	ufriio_write_request_0(sc, 0x01, 0x0000, 0x0039);
	ufriio_write_request_0(sc, 0x01, 0x0000, 0x0036);
}


//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// FIXME
// 黒凡
void ufriio_black_initialize(struct ufriio_softc *sc)
{
	// UsbProcBegin()
	ufriio_write_request_0(sc, 0x01, 0x0000, 0x0000);
	ufriio_write_request_0(sc, 0x01, 0x000f, 0x0006);
	ufriio_write_request_0(sc, 0x01, 0x0000, 0x0000);

	// UsbProcStreamInit()
	ufriio_stream_init(sc);

	// UsbProcLED(purple)
	ufriio_set_led_color(sc, UFRIIO_LED_PURPLE);

	// UsbProcFixInitB()
	ufriio_write_request_0(sc, 0x01, 0x3040, 0x0001);
	ufriio_write_request_0(sc, 0x01, 0x3030, 0x0004);
	ufriio_write_request_0(sc, 0x01, 0x3040, 0x0005);
	ufriio_write_request_0(sc, 0x01, 0x3040, 0x0007);
	ufriio_write_request_0(sc, 0x01, 0x30ff, 0x000f);
	ufriio_write_request_0(sc, 0x01, 0x3027, 0x0011);
	ufriio_write_request_0(sc, 0x01, 0x302f, 0x0013);
	ufriio_write_request_0(sc, 0x01, 0x3031, 0x0014);
	ufriio_write_request_0(sc, 0x01, 0x3002, 0x0016);
	ufriio_write_request_0(sc, 0x01, 0x3020, 0x0022);
	ufriio_write_request_0(sc, 0x01, 0x3070, 0x0024);
	ufriio_write_request_0(sc, 0x01, 0x3079, 0x002c);
	ufriio_write_request_0(sc, 0x01, 0x3034, 0x002d);
	ufriio_write_request_0(sc, 0x01, 0x3028, 0x0030);
	ufriio_write_request_0(sc, 0x01, 0x3031, 0x0031);
	ufriio_write_request_0(sc, 0x01, 0x3001, 0x0038);
	ufriio_write_request_0(sc, 0x01, 0x3078, 0x0039);
	ufriio_write_request_0(sc, 0x01, 0x3090, 0x0048);
	ufriio_write_request_0(sc, 0x01, 0x3068, 0x0051);
	ufriio_write_request_0(sc, 0x01, 0x3038, 0x005e);
	ufriio_write_request_0(sc, 0x01, 0x3021, 0x00c0);
	ufriio_write_request_0(sc, 0x01, 0x3011, 0x00c1);
	ufriio_write_request_0(sc, 0x01, 0x301a, 0x00e4);
	ufriio_write_request_0(sc, 0x01, 0x301f, 0x00ea);
}

int
ufriio_black_set_channel(struct ufriio_softc *sc, uint32_t band, uint32_t channel)
{
	// FIXME BS / CS の区別は?
	// FIXME このレンジおかしくないか? 入力値が 1 - 11 or 12 限定って何?
	// FIXME というか、チャネルなのか。周波数では無くて。
#define UFRIIO_BS 1
#define UFRIIO_CS 2
	if (UFRIIO_BS == band && 1 <= channel && channel <= 11) {
		return ufriio_black_set_channel_internal(sc, 0U, channel - 1);
	} else if (UFRIIO_CS == band && 1 <= channel && channel <= 12) {
		return ufriio_black_set_channel_internal(sc, 1U, channel - 1);
	} else {
		return (EINVAL);
	}

	sc->sc_channel = channel;
	return (0);
}

int
ufriio_black_get_signal_level(struct ufriio_softc *sc)
{
	uint32_t mix_rate;
	uint8_t sigbuf[4];

	static int af_level_table[] = {
		24,
		24,
		19,
		15,
		13,
		10,
		8,
		6,
		5,
		4,
		4,
		3,
		1,
		0,
	};

	memset(sigbuf, '\0', sizeof(sigbuf));
	ufriio_read_request_n(sc, 0x02, 0x3200, 0x00ba, sigbuf, sizeof(sigbuf));

	if (sigbuf[2] <= 0x10U) {
		return (24);
	} else if (sigbuf[2] <= 0xB0U) {
		return (0);
	} else {
		mix_rate = (uint16_t)(sigbuf[2] & 0x0fU);
		return af_level_table[sigbuf[2] >> 4] * (1 - mix_rate)
		    + af_level_table[(sigbuf[2] >> 4) + 0x01U] * mix_rate;
	}
}


//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// FIXME
// 白凡
void ufriio_white_initialize(struct ufriio_softc *sc)
{
	// UsbProcBegin()
	ufriio_write_request_0(sc, 0x01, 0x0001, 0x0000);
	ufriio_write_request_0(sc, 0x01, 0x000f, 0x0006);

	// UsbProcStreamInit()
	ufriio_stream_init(sc);

	// UsbProcLED(purple)
	ufriio_set_led_color(sc, UFRIIO_LED_PURPLE);

	// UsbProcFixInitB() / jdvbt90502_init()
	// 固定処理B
	ufriio_write_request_0(sc, 0x01, 0x3040, 0x0001);
	ufriio_write_request_0(sc, 0x01, 0x3038, 0x0004);
	ufriio_write_request_0(sc, 0x01, 0x3040, 0x0005);
	ufriio_write_request_0(sc, 0x01, 0x3040, 0x0007);
	ufriio_write_request_0(sc, 0x01, 0x304f, 0x000f);
	ufriio_write_request_0(sc, 0x01, 0x3021, 0x0011);
	ufriio_write_request_0(sc, 0x01, 0x300b, 0x0012);
	ufriio_write_request_0(sc, 0x01, 0x302f, 0x0013);
	ufriio_write_request_0(sc, 0x01, 0x3031, 0x0014);
	ufriio_write_request_0(sc, 0x01, 0x3002, 0x0016);
	ufriio_write_request_0(sc, 0x01, 0x30c4, 0x0021);
	ufriio_write_request_0(sc, 0x01, 0x3020, 0x0022);
	ufriio_write_request_0(sc, 0x01, 0x3079, 0x002c);
	ufriio_write_request_0(sc, 0x01, 0x3034, 0x002d);
	ufriio_write_request_0(sc, 0x01, 0x3000, 0x002f);
	ufriio_write_request_0(sc, 0x01, 0x3028, 0x0030);
	ufriio_write_request_0(sc, 0x01, 0x3031, 0x0031);
	ufriio_write_request_0(sc, 0x01, 0x30df, 0x0032);
	ufriio_write_request_0(sc, 0x01, 0x3001, 0x0038);
	ufriio_write_request_0(sc, 0x01, 0x3078, 0x0039);
	ufriio_write_request_0(sc, 0x01, 0x3033, 0x003b);
	ufriio_write_request_0(sc, 0x01, 0x3033, 0x003c);
	ufriio_write_request_0(sc, 0x01, 0x3090, 0x0048);
	ufriio_write_request_0(sc, 0x01, 0x3068, 0x0051);
	ufriio_write_request_0(sc, 0x01, 0x3038, 0x005e);
	ufriio_write_request_0(sc, 0x01, 0x3000, 0x0071);
	ufriio_write_request_0(sc, 0x01, 0x3008, 0x0072);
	ufriio_write_request_0(sc, 0x01, 0x3000, 0x0077);
	ufriio_write_request_0(sc, 0x01, 0x3021, 0x00c0);
	ufriio_write_request_0(sc, 0x01, 0x3010, 0x00c1);
	ufriio_write_request_0(sc, 0x01, 0x301a, 0x00e4);
	ufriio_write_request_0(sc, 0x01, 0x301f, 0x00ea);
	ufriio_write_request_0(sc, 0x01, 0x3000, 0x0077);
	ufriio_write_request_0(sc, 0x01, 0x3000, 0x0071);
	ufriio_write_request_0(sc, 0x01, 0x3000, 0x0071);
	ufriio_write_request_0(sc, 0x01, 0x300c, 0x0076);
}

int
ufriio_white_set_channel(struct ufriio_softc *sc, uint32_t channel)
{
	uint32_t wch_code = 0x0e7f + 0x002a * (channel - 13);
	uint8_t setbuf[2][5];

	if (channel < 13 || channel > 62) {
		return (EINVAL);
	}

	memset(setbuf, '\0', sizeof(setbuf));

	setbuf[0][0] = 0xc0;
	setbuf[0][1] = (uint8_t)(wch_code >> 8);
	setbuf[0][2] = (uint8_t)(wch_code & 0xff);
	setbuf[0][3] = 0xb2;
	setbuf[0][4] = 0x08;
	ufriio_write_request_n(sc, 0x03, 0x3000, 0x00fe, setbuf[0], sizeof(setbuf[0]));

	setbuf[1][0] = 0xc0;
	setbuf[1][1] = (uint8_t)(wch_code >> 8);
	setbuf[1][2] = (uint8_t)(wch_code & 0xff);
	setbuf[1][3] = 0x9a;
	setbuf[1][4] = 0x50;
	ufriio_write_request_n(sc, 0x03, 0x3000, 0x00fe, setbuf[1], sizeof(setbuf[1]));

	ufriio_read_request_1(sc, 0x02, 0x3000, 0x00b0, NULL);
	ufriio_read_request_1(sc, 0x02, 0x3000, 0x0080, NULL);

	sc->sc_channel = channel;
	return (0);
}

int
ufriio_white_get_signal_level(struct ufriio_softc *sc)
{
	// FIXME 元のコードは float で返すが、ptx(4) のインターフェイスに
	// あわせるなら、整数値にしなければいけない。
	uint32_t dw_hex;
	uint8_t sigbuf[37];

	memset(sigbuf, '\0', sizeof(sigbuf));
	ufriio_read_request_n(sc, 0x02, 0x3000, 0x0089, sigbuf, sizeof(sigbuf));

	// FIXME これってもっと汎用性のある表記にできないか?
	// エンディアンいじりだったりするとトラブルの元だが。
	dw_hex = ((uint32_t)(sigbuf[2]) << 16)
	       | ((uint32_t)(sigbuf[3]) <<  8)
	       |  (uint32_t)(sigbuf[4]);

	if (!dw_hex) {
		return (0);
	}

	/*
	 * XXX log10() / pow() cannot include with kernel.  There are libm.
	signal_level = (int)(68.4344993f
	    - log10((double)dw_hex) * 10.0
	    + pow(log10((double)dw_hex) - 5.045f, (double)4.0f) / 4.09f);
	*/
	return (dw_hex);
}

void
ufriio_set_led_color(struct ufriio_softc *sc, int color)
{
	// XXX recfriio では別のコマンドを割りあっていたが、
	// これでいいんじゃないのか?
	switch (color) {
	case UFRIIO_LED_PURPLE:
		ufriio_write_request_0(sc, 0x01, 0x000b, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000f, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0003, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0007, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000b, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000f, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0003, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0007, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0003, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0007, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000b, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000f, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0003, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0007, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000b, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000f, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000b, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000f, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0003, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0007, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000b, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000f, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000b, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000f, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000b, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000f, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000b, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000f, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000b, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000f, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000b, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000f, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000b, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000f, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000b, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000f, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0003, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0007, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0003, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0007, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0003, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0007, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0003, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0007, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0003, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0007, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0003, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0007, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0003, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0007, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0003, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0007, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000b, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000f, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000b, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000f, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000b, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000f, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000b, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000f, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000b, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000f, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000b, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000f, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000b, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000f, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000b, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000f, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0009, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000d, 0x0000);
		break;
	case UFRIIO_LED_GREEN:
		ufriio_write_request_0(sc, 0x01, 0x000b, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000f, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0003, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0007, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0003, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0007, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000b, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000f, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000b, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000f, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0003, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0007, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0003, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0007, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000b, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000f, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0003, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0007, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0003, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0007, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0003, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0007, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0003, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0007, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0003, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0007, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0003, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0007, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0003, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0007, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0003, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0007, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0003, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0007, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0003, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0007, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000b, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000f, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000b, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000f, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000b, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000f, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000b, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000f, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000b, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000f, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000b, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000f, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000b, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000f, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000b, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000f, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0003, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0007, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000b, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000f, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000b, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000f, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0003, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0007, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0003, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0007, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000b, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000f, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0003, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0007, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0003, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0007, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0001, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0005, 0x0000);
		break;
	case UFRIIO_LED_RED:
		// FIXME うまく動作してない
		ufriio_write_request_0(sc, 0x01, 0x000b, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000f, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0003, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0007, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0003, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0007, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000b, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000f, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000b, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000f, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0003, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0007, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0003, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0007, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000b, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000f, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0003, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0007, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0003, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0007, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000b, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000f, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000b, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000f, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000b, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000f, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000b, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000f, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000b, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000f, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000b, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000f, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000b, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000f, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000b, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000f, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0003, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0007, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0003, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0007, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0003, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0007, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0003, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0007, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0003, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0007, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0003, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0007, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0003, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0007, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0003, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0007, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0003, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0007, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0003, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0007, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0003, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0007, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0003, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0007, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000b, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x000f, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0003, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0007, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0001, 0x0000);
		ufriio_write_request_0(sc, 0x01, 0x0005, 0x0000);
		break;
	case UFRIIO_LED_WHITE:
		break;
	default:
		break;
	}
}

int
ufriio_black_set_channel_internal(struct ufriio_softc *sc, uint32_t space, uint32_t channel)
{
	// 資料
	// BSAT-2c/3a TsID=4[BSch番号(2byte)][TP番号]
	// BS- 1  周波数=11.727GHz 0832 TsID=4010 BS朝日
	// BS- 1  周波数=11.727GHz 0832 TsID=4011 BS-i
	// BS- 3  周波数=11.765GHz 0880 TsID=4030 WOWWOW
	// BS- 3  周波数=11.765GHz 0880 TsID=4031 BSジャパン
	// BS- 5                                   アナログで使用(WOWWOW)
	// BS- 7                                   アナログで使用(NHK BS-1)
	// BS- 9  周波数=11.880GHz 0964 TsID=4090 BS11
	// BS- 9  周波数=11.880GHz 0964 TsID=4091 スターチャンネル
	// BS- 9  周波数=11.880GHz 0964 TsID=4092 TwellV
	// BS-11                                   アナログで使用(NHK BS-2)
	// BS-13  周波数=11.957GHz 0A00 TsID=40D0 BS日テレ
	// BS-13  周波数=11.957GHz 0A00 TsID=40D1 BSフジ
	// BS-15  周波数=11.996GHz 0A4C TsID=40F1 NHK BS1/BS2/ウェザーニューズ/エンジニアリングストリーム
	// BS-15  周波数=11.996GHz 0A4C TsID=40F2 NHK BS-Hi
	// BS-17                                   未使用(BSAT-4以降に搭載予定)
	// BS-19                                   未使用(BSAT-4以降に搭載予定)
	// BS-21                                   未使用(BSAT-4以降に搭載予定)
	// BS-23                                   未使用(BSAT-4以降に搭載予定)
	//
	// NSAT-110 TsID=0000 固定
	// CSND2  周波数=12.291GHz 0C9A TsID=0000 CS
	// CSND4  周波数=12.331GHz 0CEA TsID=0000 CS
	// CSND6  周波数=12.371GHz 0D3A TsID=0000 CS
	// CSND8  周波数=12.411GHz 0D8A TsID=0000 CS
	// CSND10 周波数=12.451GHz 0DDA TsID=0000 CS
	// CSND12 周波数=12.491GHz 0E2A TsID=0000 CS
	// CSND14 周波数=12.531GHz 0E7A TsID=0000 CS
	// CSND16 周波数=12.571GHz 0ECA TsID=0000 CS
	// CSND18 周波数=12.611GHz 0F1A TsID=0000 CS
	// CSND20 周波数=12.651GHz 0F6A TsID=0000 CS
	// CSND22 周波数=12.691GHz 0FBA TsID=0000 CS
	// CSND24 周波数=12.731GHz 100A TsID=0000 CS
	
	// BS/CSch番号 第4,3桁
	static const uint8_t PllConfA[] = {
	    0x08, 0x08, 0x08, 0x08, 0x09, 0x09, 0x09, 0x0A, 0x0A, 0x0A,
	    0x0A, 0x0C, 0x0C, 0x0D, 0x0D, 0x0D, 0x0E, 0x0E, 0x0E, 0x0F,
	    0x0F, 0x0F, 0x10};   // 1/17回目 +40
	// BS/CSch番号 第2,1桁
	static const uint8_t PllConfB[] = {
	    0x32, 0x32, 0x80, 0x80, 0x64, 0x64, 0x64, 0x00, 0x00, 0x4C,
	    0x4C, 0x9A, 0xEA, 0x3A, 0x8A, 0xDA, 0x2A, 0x7A, 0xCA, 0x1A,
	    0x6A, 0xBA, 0x0A};   // 2/18回目 +40
	// TS ID 第4,3桁
	static const uint8_t PllConfC[] = {
	    0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40,
	    0x40, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	    0x00, 0x00, 0x00};   // 75回目       +02
	// TS ID 第2,1桁
	static const uint8_t PllConfD[] = {
	    0x10, 0x11, 0x30, 0x31, 0x90, 0x91, 0x92, 0xD0, 0xD1, 0xF1,
	    0xF2, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	    0x00, 0x00, 0x00};   // 76回目       +02

	struct ufriio_black_channel_sequence tmp;
	uint32_t ch_index = space * 11UL + channel;
	uint32_t xmit_num;
	uint16_t usb_value;
	uint16_t usb_index;
	uint8_t buf[3];	// FIXME ループ内で生成するのが正しいのか?
	size_t buf_len; // FIXME ループ内で生成するのが正しいのか?

	for (xmit_num = 0; ; xmit_num++) {

		tmp = ufriio_black_commands[xmit_num];
		if (tmp.req_no == 0x00) {
			break;
		}

		memset(buf, '\0', sizeof(buf));
		buf_len = 0;

		if (tmp.data_exist) {
			buf_len = 3;
			buf[0] = tmp.data1;
			buf[1] = tmp.data2;
			buf[2] = tmp.data3;
		}
		usb_value = tmp.value;
		usb_index = tmp.index;

		// 可変部の変更
		switch (xmit_num) {
		case 0:
		    buf[2] = PllConfA[ch_index];
		    break;
		case 1:
		    buf[2] = PllConfB[ch_index];
		    break;
		case 16:
		    buf[2] = PllConfA[ch_index];
		    break;
		case 17:
		    buf[2] = PllConfB[ch_index];
		    break;
		case 74:
		    // TsID1 wValueの下位Byteに入れる
		    usb_value = (usb_value & 0xff00) | (uint16_t)PllConfC[ch_index];
		    break;
		case 75:
		    // TsID2 wValueの下位Byteに入れる
		    usb_value = (usb_value & 0xff00) | (uint16_t)PllConfD[ch_index];
		    break;
		default:
		    break;
		}

		ufriio_write_request_n(sc, tmp.req_no, usb_value, usb_index, buf, buf_len);

		// 何故かここで遅延をかけないとTimeout errorが起こる。
		//usleep(SETCHANNEL_COMMAND_INTERBAL * 1000);
	}
	return (0);
}

